#coding:utf-8
'''
Базовые-и-не-только модели для подсистемы аудита

Created on 17.12.2010

@author: akvarats
'''

from django.db import models
from django.core import serializers
from django.contrib.auth.models import User, AnonymousUser

from m3.db import BaseObjectModel
from m3.core.json import json_encode
from manager import AuditManager


class BaseAuditModel(BaseObjectModel):
    '''
    Базовая модель, от которой наследуются все 
    модели хранения результатов аудита
    '''
    
    # данные пользователя. специально не делается ForeignKey.
    # чтобы не быть завязанными на ссылочную целостность
    # * логин пользователя в системе (на момент записи значения
    username = models.CharField(u'Логин пользователя', max_length=50, null=True, blank=True,
                                db_index=True, default=u'')
    
    # * идентификатор пользователя
    userid = models.PositiveIntegerField(u'Идентификатор пользователя',
                                         default=0, db_index=True)

    # * ФИО пользователя на момент записи значения (для ускоренного отображения 
    #   значений
    user_fio = models.CharField(u'ФИО пользователя', max_length=70, null=True, blank=True,
                                db_index=True, default=u'')
    
    # * дополнительные сведения о пользователе (например, сотрудником какого 
    #   учреждения он являлся на момент записи
    user_info = models.CharField(u'Дополнительные сведения о пользователе',
                                 max_length=200, null=True, blank=True, default=u'')
    
    # серверный таймстамп на запись аудита
    created = models.DateTimeField(u'Дата создания', auto_now_add=True, db_index=True)


    # список отображаемых колонок (формат - как в диктах)
    list_columns = [
        ('username', u'Пользователь'),
        ('user_info', u'Доп. инфо'),
        ('created', u'Дата'),
    ]

    # список отображаемых полей в детализации.
    # По умолчанию - все из модели аудита
    list_fields = []

    class Meta:
        abstract = True
        
    def by_user(self, user):
        '''
        Заполняет значения полей моделей на основе переданного пользователя
        '''
        if isinstance(user, User):
            self.username = user.username
            self.userid = user.id  
            self.user_fio = (user.first_name + ' ' + user.last_name).strip()
        elif isinstance(user, AnonymousUser):
            self.username = 'anonymous'
            self.userid = 0
            self.user_fio = u'<Анонимный пользователь>'

    
class BaseModelChangeAuditModel(BaseAuditModel):
    '''
    Аудит, предназначенный для отслеживания изменений в моделях системы
    '''
    
    #===========================================================================
    # Типы операций над моделью
    #===========================================================================
    ADD = 0
    EDIT = 1
    DELETE = 2
    
    # Данные модели, для которой был выполнен аудит
    # * идентификатор объекта (специально не храним FK, чтобы была возможность
    #   безболезненно удалять объекты
    object_id = models.PositiveIntegerField(u'ID записи', default=0, db_index=True)
    object_model = models.CharField(u'Модель', max_length=300, db_index=True)
    # данные модели, на момент записи аудита
    object_data = models.TextField(u'Данные модели')
    
    type = models.PositiveIntegerField(u'Действие', choices=((ADD, u'Добавление'),
                                                (EDIT, u'Изменение'),
                                                (DELETE, u'Удаление'),),
                                       db_index=True)

    list_columns = [
        ('username', u'Пользователь'),
        ('object_id', u'ID записи', 30),
        ('object_model', u'Модель'),
        ('type', u'Действие', 70),
        ('created', u'Дата'),
    ]
    
    @classmethod
    def write(cls, user, model_object, type, *args, **kwargs):
        
        audit = cls()
        audit.by_user(user)
        
        if (type == BaseModelChangeAuditModel.ADD or 
            type == 'add' or type == 'new'):
            audit.type = BaseModelChangeAuditModel.ADD
            
        elif (type == BaseModelChangeAuditModel.EDIT or
              type == 'edit'):
            audit.type = BaseModelChangeAuditModel.EDIT
            
        elif (type == BaseModelChangeAuditModel.DELETE or
              type == 'remove' or
              type == 'delete'):
            audit.type = BaseModelChangeAuditModel.DELETE
            
        audit.object_id = model_object.id if model_object.id else 0
        audit.object_model = (model_object.__class__.__module__ + 
                              '.' + 
                              model_object.__class__.__name__)
        try:
            audit.object_data = serializers.serialize('json', [model_object,])
        except:
            pass
        audit.save()
    
    class Meta:
        abstract = True


#===============================================================================
# Общий аудит для сохранения информации об изменении моделей
#===============================================================================
class DefaultModelChangeAuditModel(BaseModelChangeAuditModel):
    '''
    Модель дефолтного аудита изменения моделей
    '''
    
    class Meta:
        verbose_name = u'Изменения таблиц системы'
        db_table = 'm3_audit_model_changes'
        
AuditManager().register('model-changes', DefaultModelChangeAuditModel)        
#===============================================================================
# Аудит для сохранения информации об изменении записей справочников
#===============================================================================
class DictChangesAuditModel(BaseModelChangeAuditModel):
    '''
    Модель аудита изменения в справочниках
    '''
    
    class Meta:
        verbose_name = u'Изменения в справочниках'
        db_table = 'm3_audit_dict_changes'

AuditManager().register('dict-changes', DictChangesAuditModel) 
#===============================================================================
# Преднастроенный аудит для входов/выходов пользователей из системы
#===============================================================================
class AuthAuditModel(BaseAuditModel):
    '''
    Аудит входов/выходов пользователя из системы
    '''
    #===========================================================================
    # Тип авторизации пользователя
    #===========================================================================
    LOGIN = 0
    LOGOUT = 1
    
    type = models.PositiveIntegerField(u'Действие', choices=((LOGIN, u'Вход в систему'),
                                                (LOGOUT, u'Выход из системы'),), 
                                       db_index=True)

    list_columns = [
        ('userid', u'ID', 30),
        ('username', u'Пользователь'),
        ('type', u'Действие'),
        ('created', u'Дата'),
    ]

    @classmethod
    def write(cls, user, type='login', *args, **kwargs):
        '''
        Пишем информацию об аудите входа/выхода
        '''
        audit = AuthAuditModel()
        audit.by_user(user)
        audit.type = AuthAuditModel.LOGIN if type == 'login' else AuthAuditModel.LOGOUT
        audit.save()
    
    class Meta:
        verbose_name = u'Авторизации пользователей'
        db_table = 'm3_audit_auth'

AuditManager().register('auth', AuthAuditModel)

## Структура данных аудита по действиям над ролями пользователей
class RolesAuditModel(BaseAuditModel):
    ''' Структура данных аудита действий над ролями пользователей '''
    PERMISSION_ADDITION = 0
    PERMISSION_REMOVAL = 1
    PERMISSION_ENABLEMENT = 2
    PERMISSION_DISABLEMENT = 3
    
    type = models.PositiveIntegerField(u'Действие', choices=((PERMISSION_ADDITION, u'Добавление прав'),
                                                (PERMISSION_REMOVAL, u'Лишение прав'),
                                                (PERMISSION_ENABLEMENT, u'Активация права'),
                                                (PERMISSION_DISABLEMENT, u'Отключение права')
                                                ))
    role_id = models.PositiveIntegerField(u'ID роли')
    role_name = models.CharField(u'Роль', max_length=200)
    permission_code = models.CharField(u'Разрешения', max_length=200,
                                        null=True, blank=True, default='')

    list_columns = [
        ('username', u'Пользователь'),
        ('role_id', u'ID роли', 30),
        ('role_name', u'Роль'),
        ('type', u'Действие', 70),
        ('created', u'Дата'),
    ]

    @classmethod
    def write(cls, user, role, permission_or_code=None, type=PERMISSION_ADDITION,
               *args, **kwargs):
        ''' Непосредственная запись '''
        audit = RolesAuditModel(); audit.by_user(user)
        
        audit.type = type
        audit.role = role
        audit.permission_code = cls.parse_permission(permission_or_code)
        audit.save()
    
    def get_role(self):
        return RolesAuditModel.objects.get(self.role_id)
    def set_role(self, role):
        self.role_id = role.id
        self.role_name = role.name
    role = property(get_role, set_role)
    
    @classmethod
    def parse_permission(cls, perm):    
        permission_code = perm; return permission_code
    
    class Meta:
        verbose_name = u'Изменения ролей пользователей'
        db_table = 'm3_audit_roles' 
        
AuditManager().register('roles', RolesAuditModel)


#coding:utf-8

import os

project_path = os.path.dirname(__file__)

__version__ = '1.0'
__appname__ = u'Тестовое приложение для m3_smthng'

__require__ = {
    'm3': '0.9',
}

# исходные тексты m3_smthng всегда забираем локально из текущего репозитория
# поэтому путь такой вот
__require_local__ = {
    'm3_smthng': os.path.realpath(os.path.join(project_path, '../../')),
}



#coding:utf-8
'''
Модуль, который выполняет подготовку 
'''
import sys
import os
import Queue
import subprocess
import types
import datetime
import shutil
from decimal import Decimal


REPO_LOCATION = 'https://readonly:onlyread@repos.med.bars-open.ru/'
TOPMOST_VERSION = 'zzzzzzzzz'

#===============================================================================
# Точка входа.
#===============================================================================

def start(as_repos=False):
    '''
    Точка входа. 
    '''
    has_error= False
    
    script_root = os.path.dirname(__file__)
    env_root = os.path.join(script_root, '../env')
    envbuild_root = os.path.join(env_root, '.build')
    
    # достаем версию и зависимости основного продукта
    module = import_module('version')
    
    # словарь уже обработанных приложений
    app_queue = Queue.Queue()  
    processed_apps = {}
    synced_locally = {} 
    
    root_version, root_appname, root_require, root_require_locals = app_version('')
    
    print u'--------------------------------------------------------------------'
    print u'Сборка окружения для проекта \'%s\', версия %s' % (root_appname, root_version)
    print u'--------------------------------------------------------------------'
    
    # создаем папки для хранения информации
    if not os.path.exists(env_root):
        os.mkdir(env_root)
        
    #sys.path.insert(0, env_root)
    sys.path.insert(0, os.path.dirname(__file__))
    
        
    if not os.path.exists(envbuild_root):
        os.mkdir(envbuild_root)
    
    # забираем зависимости корневого приложения
    for require_app, require_version in root_require.iteritems():
        app_queue.put(AppDescription(require_app, require_version))
        
    for require_app, require_path in root_require_locals.iteritems():
        app_queue.put(AppDescription(require_app, require_path, is_local=True))
        
    
    #os.chdir('../')
    # основной цикл обработки
    while not app_queue.empty():
        
        # забираем приложение из очереди на обработку
        app = app_queue.get_nowait()
        print '>>', app.name, '(' + app.version + ')'
        
        if app.is_local:
            
            print '  ', u'Выполняется ЛОКАЛЬНАЯ синхронизация исходных текстов приложения', app.name
            
            # данное приложение необходимо синхронизировать только локально
            # поэтому тупо копируем данные
            # путь к исходным текстам по счастливой случайности хранится
            # в app.version
            
            if not os.path.exists(app.version):
                print '!!', u'Папка исходных текстов', app.version, u'не найдена'
                has_error = True
                continue
            
            copy_sources(app.name, env_root, app.version)
            
            # отмечаем, что данное приложение имеет наивысшую версию
            # и не может быть замещено ни одиним другим набором исходников
            # из репозитариев или других папок
            processed_apps[app.name] = TOPMOST_VERSION
            synced_locally[app.name] = app.version
            
        else:
            
            if processed_apps.get(app.name, '.') >= app.version:
                # точка всегда "меньше" любого другого символа
                print '  ', u'Исходные тексты', app.name, u'находятся в актуальном состоянии'
                continue
            # подготовка приложения
            app_repo_root = os.path.join(envbuild_root, app.name) 
            if not os.path.exists(app_repo_root):
                print '  ', u'Клонирование репозитария приложения', app.name
                if not clone_repo(app.name, app_repo_root):
                    has_error = True
                    print '!!', u'Исходные тексты приложения', app.name, u'из-за возникшей ошибки находятся в неактуальном состоянии'
                    continue
            else:
                print '  ', u'Пул репозитария приложения', app.name
                if not pull_repo(app_repo_root):
                    has_error = True
                    print '!!', u'Исходные тексты приложения', app.name, u'из-за возникшей ошибки находятся в неактуальном состоянии'
                    continue
            
            print '  ', u'Обновление репозитария приложения', app.name, u'на ветку', app.version
            if not update_repo(app_repo_root, app.version):
                has_error = True
                print '!!', u'Исходные тексты приложения', app.name, u'из-за возникшей ошибки находятся в неактуальном состоянии'
                continue
        
            # копируем получившееся в ./env
            print '  ', u'Копирование исходных кодов приложения в env'    
            copy_sources(app.name, env_root, app_repo_root)
            
            processed_apps[app.name] = app.version
            
        # производим разбор
        app_source_path = os.path.join(env_root, app.name)
        if os.path.exists(os.path.join(app_source_path, 'version.py')):
            sys.path.insert(0, app_source_path)
            _, _, require, require_locals = app_version()
            
            for require_app, require_version in require.iteritems():
                print '  ', app.name, 'требует приложение', require_app, u'версии', require_version
                app_queue.put(AppDescription(require_app, require_version))
    
    print u'--------------------------------------------------------------------'
    if has_error:
        print u'Сборка окружения завершена С ОШИБКАМИ.'
    else:
        print u'Сборка окружения успешно завершена'
        
    print u'--------------------------------------------------------------------'
    print u'Выполнена синхронизация следующих пакетов:'
    print '  ', root_appname + ':', root_version
    for app, version in processed_apps.iteritems():
        print '  ', app + ':', synced_locally.get(app, version)
    
def clone_repo(app_name, app_repo_root):
    
    out, err = run_command(['hg', 'clone', REPO_LOCATION + app_name, app_repo_root])
    if err:
        print '  ', u'Клонирование репозитория', REPO_LOCATION + app_name, u'завершено с ошибкой:', err
        
    return not err

def pull_repo(app_repo_root):
    
    out, err = run_command(['hg', 'pull', '-R', app_repo_root])
    if err:
        print '  ', u'Пул репозитория', app_repo_root, u'завершен с ошибкой:', err
        
    return not err

def update_repo(app_repo_root, app_version):
    
    out, err = run_command(['hg', 'update', app_version, '-R', app_repo_root])
    if err:
        print '  ', u'Обновление репозитория', app_repo_root, u'не ветку', app_version, u'завершено с ошибкой:', err
        
    return not err

def copy_sources(app_name, env_root, app_repo_root):
    
    app_source_root = os.path.join(env_root, app_name)
    if os.path.exists(app_source_root):
        shutil.rmtree(path=app_source_root)
    
    shutil.copytree(src=os.path.join(app_repo_root, 'src/' + app_name), dst=app_source_root)
    

def app_version(app_name=''):    
    '''
    Возвращает кортеж из трех элементов (app_version, require, require_local)
    '''
    module = import_module('%s.version' % app_name if app_name else 'version' )
    reload(module)
        
    return (module.__version__,
            module.__appname__,
            module.__require__,
            module.__require_local__ if hasattr(module, '__require_local__') else {},)
    
#===============================================================================
# Внутренние классы
#===============================================================================
class AppDescription(object):
    
    def __init__(self, name='', version='', is_local = False):
        self.name = name
        self.version = version
        self.is_local = is_local

#===============================================================================
# Внутренние функции
#===============================================================================
def run_command(command):
    '''
    Выполняет команду как subprocess,
    возвращает (stdout, stderr) процесса 
    '''
    popen = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    popen.wait()
    return popen.communicate()


def _resolve_name(name, package, level):
    """Return the absolute name of the module to be imported."""
    if not hasattr(package, 'rindex'):
        raise ValueError("'package' not set to a string")
    dot = len(package)
    for x in xrange(level, 1, -1):
        try:
            dot = package.rindex('.', 0, dot)
        except ValueError:
            raise ValueError("attempted relative import beyond top-level "
                              "package")
    return "%s.%s" % (package[:dot], name)


def import_module(name, package=None):
    """Import a module.

    The 'package' argument is required when performing a relative import. It
    specifies the package to use as the anchor point from which to resolve the
    relative import to an absolute import.

    """
    if name.startswith('.'):
        if not package:
            raise TypeError("relative imports require the 'package' argument")
        level = 0
        for character in name:
            if character != '.':
                break
            level += 1
        name = _resolve_name(name[level:], package, level)
    __import__(name)
    return sys.modules[name]

def force_unicode(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Similar to smart_unicode, except that lazy instances are resolved to
    strings, rather than kept as lazy objects.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    # Handle the common case first, saves 30-40% in performance when s
    # is an instance of unicode. This function gets called often in that
    # setting.
    if isinstance(s, unicode):
        return s
    if strings_only and is_protected_type(s):
        return s
    try:
        if not isinstance(s, basestring,):
            if hasattr(s, '__unicode__'):
                s = unicode(s)
            else:
                try:
                    s = unicode(str(s), encoding, errors)
                except UnicodeEncodeError:
                    if not isinstance(s, Exception):
                        raise
                    # If we get to here, the caller has passed in an Exception
                    # subclass populated with non-ASCII data without special
                    # handling to display as a string. We need to handle this
                    # without raising a further exception. We do an
                    # approximation to what the Exception's standard str()
                    # output should be.
                    s = ' '.join([force_unicode(arg, encoding, strings_only,
                            errors) for arg in s])
        elif not isinstance(s, unicode):
            # Note: We use .decode() here, instead of unicode(s, encoding,
            # errors), so that if s is a SafeString, it ends up being a
            # SafeUnicode at the end.
            s = s.decode(encoding, errors)
    except UnicodeDecodeError, e:
        if isinstance(s, Exception):
            
            # If we get to here, the caller has passed in an Exception
            # subclass populated with non-ASCII bytestring data without a
            # working unicode method. Try to handle this without raising a
            # further exception by individually forcing the exception args
            # to unicode.
            s = ' '.join([force_unicode(arg, encoding, strings_only,
                    errors) for arg in s])
    return s

def is_protected_type(obj):
    """Determine if the object instance is of a protected type.

    Objects of protected types are preserved as-is when passed to
    force_unicode(strings_only=True).
    """
    return isinstance(obj, (
        types.NoneType,
        int, long,
        datetime.datetime, datetime.date, datetime.time,
        float, Decimal)
    )


if __name__ == '__main__':
    start()



#coding:utf-8
'''
Created on 22.12.2010

@author: akvarats
'''

class M3AuditException(Exception):
    '''
    Некоторое исключение подсистемы аудита
    '''
    pass

class DropM3AuditCacheException(M3AuditException):
    '''
    Исключение, которое возникает при попытке сброса кеша, хранящего
    зарегистрированные типы аудита
    '''
    pass

class NoWriteMethonInM3AuditException(M3AuditException):
    '''
    Выбрасывается в случае, если при регистрации аудита у класса модели
    не найден 
    '''
    pass


#coding:utf-8
'''
Пакет тестов для Платформы М3.

Тесты запускаются командой

* все тесты: 
  python manage.py test
  
* тесты конкретного приложения: 
  python manage.py test tests.data.caching_tests

* тесты нескольких конкретных приложений:
  python manage.py test tests.data.caching_tests tests.data.mie_tests
'''




#coding:utf-8
'''
Created on 06.01.2011

@author: akvarats
'''
from m3.ui.actions import ActionPack, Action, utils, ControllerCache
from m3.helpers import urls
from m3.ui import actions

import ui
from manager import AuditManager

from datetime import datetime, time

class BaseAuditUIActions(actions.ActionPack):
    '''
    Базовый пакет действий, который позволяет
    '''
    
    model = None
    acd = []

    # список отображаемых аудитов в комбо. По умолчанию - все
    list_audits = AuditManager().list()
    
    def __init__(self):
        super(BaseAuditUIActions, self).__init__()

        self.list_window_action = BaseAuditListWindowAction()
        self.rows_action = AuditRowsDataAction()
        self.fields_action = AuditRowFieldsDataAction()
        self.list_sort_order = ['-created']
        
        self.actions.extend([self.rows_action,
                             self.list_window_action,
                             self.fields_action,
                             ])

    def get_acd(self):
        '''
        Возвращает context declaration на основе данных, прописанных
        при настройке дочернего класса
        '''
        if hasattr(self, 'acd'):
            return self.acd
    
    def get_list_window(self, window, request, context):
        '''
        Возвращает окно, в котором отображаются результаты аудита
        '''
        return window

    def get_list_url(self):
        '''
        Возвращает полный URL до действия на получение окна со списком
        записей аудита
        '''
        return self.list_window_action.absolute_url()

    def get_rows_url(self):
        '''
        Возвращает полный URL до действия на получение записей аудита
        '''
        return self.rows_action.absolute_url()

    def get_row_fields_url(self):
        '''
        Возвращает полный URL до действия на получение данных по записи аудита
        '''
        return self.fields_action.absolute_url()

    def configure_window(self, window, request, context):
        '''
        Конфигурация окна, выделено для возможности переопределения
        '''
        top_panel = self.get_top_panel(window, request, context)
        if top_panel:
            window.items.append(top_panel)
        
        east_panel = self.get_east_panel(window, request, context)
        if east_panel:
            window.items.append(east_panel)
        return window

    def get_east_panel(self, window, request, context):
        '''
        Возвращает панель с опциональными элементами формы,
        построенными по модели (list_fields)
        '''
        panel = ui.DefaultEastPanel()
        model = AuditManager().get(context.audit)
        panel.create_fields(model)
        window.url_audit_row = self.get_row_fields_url()
        window.form_details = panel.form

        return panel

    def get_top_panel(self, window, request, context):
        '''
        Возвращает панель с опциональным списком для выбора аудита (list_audit)
        '''
        panel = ui.DefaultTopPanel()

        # формируем список аудитов для комбо
        audits = []
        for audit in self.list_audits:
            model = AuditManager().get(audit)
            audits.append((audit, model.get_verbose_name()))

        panel.create_audits_combo(audits, context.audit)
        window.url_audit_win = self.get_list_url()

        return panel

    def apply_column_filter(self, pre_query, request, context):
        '''
        Накладывает на запрос колоночный фильтр
        '''
        columns_map = {}
        model = pre_query.model
        for column in model.list_columns:
            if column[0] != 'created':
                columns_map[column[0]] = column[0]

        pre_query = utils.apply_column_filter(pre_query, request, columns_map)
        return pre_query

    def get_rows(self, pre_query, request, context):
        '''
        '''
        return pre_query


class BaseAuditListWindowAction(actions.Action):
    '''
    Получение окна показа списка записей
    '''
    url = '/list-window'
    shortname = 'audit-list-window'

    def __init__(self, *args, **kwargs):
        super(BaseAuditListWindowAction, self).__init__(*args, **kwargs)

    def context_declaration(self):
        acd_list = self.parent.get_acd()
        acd_list.extend([actions.ACD(name='audit', type=str, required=True, default=''),])
        return acd_list

    def run(self, request, context):
        if not context.audit:
            # По умолчанию откроем первый попавшийся аудит
            context.audit = self.parent.list_audits[0]

        window = ui.AuditListWindow()
        window = self.parent.get_list_window(window, request, context)

        model = AuditManager().get(context.audit)
        window.create_columns(model.list_columns, model)
        window.grid_rows.url_data = self.parent.get_rows_url()
        window = self.parent.configure_window(window, request, context)

        return actions.ExtUIScriptResult(data=window)


class AuditRowsDataAction(actions.Action):
    '''
    Действие на получение списка записей выбранного аудита
    '''
    url = '/rows'
    shortname = 'audit-rows'

    def context_declaration(self):
        acd_list = self.parent.get_acd()
        acd_list.extend([actions.ACD(name='audit', type=str, required=True),
                         actions.ACD(name='start', type=int, required=True, default=-1),
                         actions.ACD(name='limit', type=int, required=True, default=-1),
                         actions.ACD(name='created', type=datetime)])
        return acd_list

    def run(self, request, context):
        '''
        Выполняет построение предварительного запроса и отдает в функцию
        '''
        direction = request.REQUEST.get('dir')
        user_sort = request.REQUEST.get('sort')
        if direction == 'DESC':
            user_sort = '-' + user_sort
        sort_order = [user_sort] if user_sort else self.parent.list_sort_order

        model = AuditManager().get(context.audit)
        pre_query = model.objects.all()

        if request.REQUEST.get('created'):
           end_day = datetime.combine(context.created.date(), time.max)
           pre_query = pre_query.filter(created__gte=context.created)
           pre_query = pre_query.filter(created__lte=end_day)

        pre_query = self.parent.apply_column_filter(pre_query, request, context)
        pre_query = utils.apply_sort_order(pre_query, model.list_columns, sort_order)

        query = self.parent.get_rows(pre_query, request, context)

        return actions.ExtGridDataQueryResult(data=query,
                                              start=context.start,
                                              limit=context.limit)


class AuditRowFieldsDataAction(actions.Action):
    '''
    Действие на получение записи аудита
    '''
    url = '/row-fields'
    shortname = 'audit-row-fields'

    def context_declaration(self):
        acd_list = self.parent.get_acd()
        acd_list.extend([actions.ACD(name='audit', type=str, required=True),
                         actions.ACD(name='id', type=int, required=True, default=-1),])
        return acd_list

    def run(self, request, context):
        model = AuditManager().get(context.audit)
        data = model.objects.get(pk=context.id)

        return actions.PreJsonResult({'data': data})


#coding: utf-8
'''
Приложение, реализующее журналирование действий в m3 и прикладном приложении.

Предполагаемый сценарий использования:
...

Для вызова преднастроенного окна со списком нескольких или одного аудита необходимо
отнаследовать базовый экшнпак аудитов BaseAuditUIActions, где задать в list_audits
необходимые для отображения аудиты
'''

from models import BaseAuditModel
from exceptions import M3AuditException
from manager import AuditManager


#coding: utf-8
"""
Created on 29.10.11
@author: ruldashev
"""
from django.db.models.base import ModelBase
from django import forms as form_fields
from django.utils.datastructures import SortedDict

from m3.ui.ext import fields as ext_fields
from m3.ui.ext.misc import store


def fields_for_model(model, fields=None, exclude=None, widgets=None, formfield_callback=None):
    """
    Returns a ``SortedDict`` containing form fields for the given model.

    ``fields`` is an optional list of field names. If provided, only the named
    fields will be included in the returned fields.

    ``exclude`` is an optional list of field names. If provided, the named
    fields will be excluded from the returned fields, even if they are listed
    in the ``fields`` argument.

    Вынужденная копия django.forms.models.fields_for_model с тем лишь отличием,
    что мы вытаскиваем также non editable поля
    """
    field_list = []
    ignored = []
    opts = model._meta
    for f in opts.fields + opts.many_to_many:
        if fields is not None and not f.name in fields:
            continue
        if exclude and f.name in exclude:
            continue
        if widgets and f.name in widgets:
            kwargs = {'widget': widgets[f.name]}
        else:
            kwargs = {}

        if formfield_callback is None:
            formfield = f.formfield(**kwargs)
        elif not callable(formfield_callback):
            raise TypeError('formfield_callback must be a function or callable')
        else:
            formfield = formfield_callback(f, **kwargs)

        if formfield:
            field_list.append((f.name, formfield))
        else:
            ignored.append(f.name)
    field_dict = SortedDict(field_list)
    if fields:
        field_dict = SortedDict(
            [(f, field_dict.get(f)) for f in fields
                if ((not exclude) or (exclude and f not in exclude)) and (f not in ignored)]
        )
    return field_dict


def ext_fields_for_model(model, exclusion=[]):
    '''
    Возвращает список ExtFields для полей модели
    '''
    assert isinstance(model, ModelBase)
    model_ext_fields = {}
    
    # маппер Django Forms Fields -> ExtFields
    # TODO: использовать общий маппер (если/когда он будет)
    mapper = dict()
    default_field = ext_fields.ExtStringField
    mapper[form_fields.IntegerField] = ext_fields.ExtNumberField
    mapper[form_fields.Textarea] = ext_fields.ExtTextArea
    mapper[form_fields.DateTimeInput] = ext_fields.ExtDateField
    mapper[form_fields.Select] = ext_fields.ExtComboBox

    fields = fields_for_model(model).items()
    for key, field in fields:
        if key in exclusion:
            continue
        
        field_cls = field.widget.__class__
        ext_field = mapper[field_cls]() if field_cls in mapper else default_field()

        ext_field.name = key
        ext_field.label = field.label
        if hasattr(field, 'choices'):
            # если есть список, значит построим комбобокс
            ext_field.set_store(store.ExtDataStore(data=field.choices))
            ext_field.display_field = 'name'
            ext_field.value_field = 'id'
            ext_field.trigger_action = 'all'

        model_ext_fields[key] = ext_field
        
    return model_ext_fields


#coding:utf-8
'''

Описание менеджера, который обслуживает систему аудита

Created on 21.12.2010

@author: akvarats
'''

from django.db import transaction

from m3.helpers import logger
from m3.data.caching import RuntimeCache

from exceptions import (DropM3AuditCacheException, 
                        NoWriteMethonInM3AuditException)


__all__=('AuditManager',)


class AuditCache(RuntimeCache):
    '''
    Кеш с информацией о зарегистрированных типах аудита
    '''
    def drop(self, dimensions):
        raise DropM3AuditCacheException()
    
    def drop_all(self):
        raise DropM3AuditCacheException()
    
    

class AuditManager(object):
    '''
    Менеджер, который выполняет запись результатов аудита
    '''
    
    def register(self, audit_name, audit_model):
        '''
        Регистрирует тип аудита
        '''
        
        if not hasattr(audit_model, 'write'):
            raise NoWriteMethonInM3AuditException(audit_model.__module__ + '.' + audit_model.__name__)
        
        AuditCache().set(audit_name, audit_model)

    def list(self):
        return [x[0] for x in AuditCache().data]
        
    def get(self, audit_name, default=None):
        '''
        Возвращает данные, хранящиеся для аудита с указанным именем.
        
        Если аудит не найден, возвращается значение параметра default. 
        '''
        return AuditCache().get(audit_name, default)
    
    @transaction.commit_on_success
    def write(self, audit_name, user, *args, **kwargs):
        '''
        Выполняет операцию записи об аудите в базу данных
        '''
        audit = self.get(audit_name, None)
        if audit:
            try:
                audit.write(user, *args, **kwargs)
            except:
                logger.exception(u'Не удалось записать результаты аудита \'%s\'' % audit_name)



#coding:utf-8
'''
Created on 06.01.2011

@author: akvarats
'''
from m3.ui.ext import windows
from m3.ui.ext import panels
from m3.ui.ext import containers
from m3.ui.ext import fields
from m3.ui.ext.misc import store

from manager import AuditManager
from helpers import ext_fields_for_model


class AuditListWindow(windows.ExtWindow):
    '''
    Окно со списком записей аудита
    '''
    def __init__(self, *args, **kwargs):
        super(AuditListWindow, self).__init__(*args, **kwargs)

        self.title = u'Аудит операций'
        self.template_globals = 'm3-audit-list-window.js'

        self.width = 800
        self.height = 550
        self.layout = 'border'
        self.maximizable = True

        self.panel_center = containers.ExtContainer()
        self.panel_center.region = 'center'
        self.panel_center.layout = 'fit'

        self.grid_rows = panels.ExtObjectGrid()
        self.grid_rows.sm = containers.ExtGridRowSelModel(single_select=True)
        self.grid_rows.store.remote_sort = True
        self.grid_rows.handler_click = 'rowChangeHandler'
        self.grid_rows.plugins.extend([
            'new Ext.ux.grid.GridHeaderFilters()',
        ])
        
        self.panel_center.items.append(self.grid_rows)
        self.items.extend([self.panel_center,])

    def create_columns(self, columns, model=None):
        '''
        Добавляет отображаемые колонки. Реализазация - как в диктах
        '''

        # получим поля модели в виде ExtFields,
        # чтобы сформировать подходящие хэдэр-фильтры
        model_ext_fields = ext_fields_for_model(model)

        for column in columns:
            if isinstance(column, tuple):
                column_params = { 'data_index': column[0], 'header': column[1],
                                  'sortable': True }
                if len(column)>2:
                    column_params['width'] = column[2]
            elif isinstance(column, dict):
                column_params = column
            else:
                raise Exception('Incorrect parameter column.')

            if column[0] in model_ext_fields:
                ext_field = model_ext_fields[column[0]]
                column_params['extra'] = {
                    'filter': ext_field.render(),
                }
                # для enum-колонок сформируем рендерер, ктр будет заменять id
                # на человеческие значения
                if hasattr(ext_field, 'store'):
                    renderer = u'function(v){var store={%s};return store[v]}' %\
                               ','.join('"%s":"%s"' % (key, val) for key, val in ext_field.store.data)
                    column_params['extra']['renderer'] = renderer

            self.grid_rows.add_column(**column_params)

        
class DefaultEastPanel(panels.ExtPanel):
    '''
    Панель с полями конкретной записи
    '''
    
    def __init__(self, *args, **kwargs):
        super(DefaultEastPanel, self).__init__(*args, **kwargs)

        self.region = 'east'
        self.layout = 'fit'
        self.title = u'Запись аудита'
        self.width = 270
        self.min_width = 200
        self.max_width = 500
        self.split = True
        self.collapsible = True
        self.padding = 5
        self.body_cls = 'x-window-mc'
        self.template_globals = 'm3-audit-list-window.js'

        self.form = containers.ExtForm()
        self.items.append(self.form)
        
        self.init_component(*args, **kwargs)

    def create_fields(self, model):
        '''
        Добавляет поля для просмотра записи на основе полей в модели
        '''
        ext_fields = ext_fields_for_model(model)

        if model.list_fields:
            # если указаны кастомные поля для отображения, то добавляем на форму их
            list_fields = model.list_fields
        else:
            # в противном случае - все поля модели
            list_fields = ext_fields.keys()

        for field_name in list_fields:
            if field_name in ext_fields:
                ext_field = ext_fields[field_name]
                ext_field.read_only = True
                ext_field.anchor = '100%'

                self.form.items.append(ext_field)


class DefaultTopPanel(containers.ExtContainer):
    '''
    Панель с выбором нужного аудита
    '''

    def __init__(self, *args, **kwargs):
        super(DefaultTopPanel, self).__init__(*args, **kwargs)

        self.region = 'north'
        self.layout = 'form'
        self.height = 35
        self.style = {'padding': '5px'}

        self.init_component(*args, **kwargs)
    
    def create_audits_combo(self, list_audits, current_audit=''):
        '''
        Создает комбо со списком аудитов для перехода между ними
        '''
        f_audits_combo = fields.ExtComboBox()
        f_audits_combo.name = 'audit_id'
        f_audits_combo.label = u'Аудит'
        f_audits_combo.value = current_audit
        f_audits_combo.value_field = 'id'
        f_audits_combo.display_field = 'name'
        f_audits_combo.trigger_action = 'all'
        f_audits_combo.anchor = '50%'
        f_audits_combo.editable = False
        f_audits_combo.allow_blank = False
        f_audits_combo.handler_select = 'changeAuditHandler'
        f_audits_combo.set_store(store.ExtDataStore(data=list_audits))

        self.items.append(f_audits_combo)


#coding:utf-8
'''
Created on 07.01.2011

@author: akvarats
'''

#===============================================================================
# js shortcuts
#===============================================================================

def js_audit_list_window(audit_pack, parent_window):
    '''
    Возвращает js код, который может быть использован для показа окна
    с результатами аудита
    '''
    js_template = '''function (){{
    var parent_window = Ext.getCmp('{parent_window_client_id}');
    var params = {{}};
    Ext.applyIf(params, parent_window.actionContextJson);
    var loadMask = new Ext.LoadMask(win.body);
    loadMask.show();
    Ext.Ajax.request({{
        method: 'POST',
        url: '{action_url}',
        params: params,
        success: function(response, opts){{
           loadMask.hide();
           smart_eval(response.responseText); 
        }},
        failure: function (response, opts){{
            loadMask.hide();
            uiAjaxFailMessage.apply(this, arguments);
        }}
    }});
}}'''

    return js_template.format(parent_window_client_id=parent_window.client_id,
                              action_url=audit_pack.get_list_url())


# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'DefaultModelChangeAuditModel'
        db.create_table('m3_audit_model_changes', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('username', self.gf('django.db.models.fields.CharField')(default=u'', max_length=50, null=True, db_index=True, blank=True)),
            ('userid', self.gf('django.db.models.fields.PositiveIntegerField')(default=0, db_index=True)),
            ('user_fio', self.gf('django.db.models.fields.CharField')(default=u'', max_length=70, null=True, db_index=True, blank=True)),
            ('user_info', self.gf('django.db.models.fields.CharField')(default=u'', max_length=200, null=True, blank=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, db_index=True, blank=True)),
            ('object_id', self.gf('django.db.models.fields.PositiveIntegerField')(default=0, db_index=True)),
            ('object_model', self.gf('django.db.models.fields.CharField')(max_length=300, db_index=True)),
            ('object_data', self.gf('django.db.models.fields.TextField')()),
            ('type', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
        ))
        db.send_create_signal('m3_audit', ['DefaultModelChangeAuditModel'])

        # Adding model 'DictChangesAuditModel'
        db.create_table('m3_audit_dict_changes', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('username', self.gf('django.db.models.fields.CharField')(default=u'', max_length=50, null=True, db_index=True, blank=True)),
            ('userid', self.gf('django.db.models.fields.PositiveIntegerField')(default=0, db_index=True)),
            ('user_fio', self.gf('django.db.models.fields.CharField')(default=u'', max_length=70, null=True, db_index=True, blank=True)),
            ('user_info', self.gf('django.db.models.fields.CharField')(default=u'', max_length=200, null=True, blank=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, db_index=True, blank=True)),
            ('object_id', self.gf('django.db.models.fields.PositiveIntegerField')(default=0, db_index=True)),
            ('object_model', self.gf('django.db.models.fields.CharField')(max_length=300, db_index=True)),
            ('object_data', self.gf('django.db.models.fields.TextField')()),
            ('type', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
        ))
        db.send_create_signal('m3_audit', ['DictChangesAuditModel'])

        # Adding model 'AuthAuditModel'
        db.create_table('m3_audit_auth', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('username', self.gf('django.db.models.fields.CharField')(default=u'', max_length=50, null=True, db_index=True, blank=True)),
            ('userid', self.gf('django.db.models.fields.PositiveIntegerField')(default=0, db_index=True)),
            ('user_fio', self.gf('django.db.models.fields.CharField')(default=u'', max_length=70, null=True, db_index=True, blank=True)),
            ('user_info', self.gf('django.db.models.fields.CharField')(default=u'', max_length=200, null=True, blank=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, db_index=True, blank=True)),
            ('type', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
        ))
        db.send_create_signal('m3_audit', ['AuthAuditModel'])


    def backwards(self, orm):
        
        # Deleting model 'DefaultModelChangeAuditModel'
        db.delete_table('m3_audit_model_changes')

        # Deleting model 'DictChangesAuditModel'
        db.delete_table('m3_audit_dict_changes')

        # Deleting model 'AuthAuditModel'
        db.delete_table('m3_audit_auth')


    models = {
        'm3_audit.authauditmodel': {
            'Meta': {'object_name': 'AuthAuditModel', 'db_table': "'m3_audit_auth'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'type': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'user_fio': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '70', 'null': 'True', 'db_index': 'True', 'blank': 'True'}),
            'user_info': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '200', 'null': 'True', 'blank': 'True'}),
            'userid': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'db_index': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '50', 'null': 'True', 'db_index': 'True', 'blank': 'True'})
        },
        'm3_audit.defaultmodelchangeauditmodel': {
            'Meta': {'object_name': 'DefaultModelChangeAuditModel', 'db_table': "'m3_audit_model_changes'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'object_data': ('django.db.models.fields.TextField', [], {}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'db_index': 'True'}),
            'object_model': ('django.db.models.fields.CharField', [], {'max_length': '300', 'db_index': 'True'}),
            'type': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'user_fio': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '70', 'null': 'True', 'db_index': 'True', 'blank': 'True'}),
            'user_info': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '200', 'null': 'True', 'blank': 'True'}),
            'userid': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'db_index': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '50', 'null': 'True', 'db_index': 'True', 'blank': 'True'})
        },
        'm3_audit.dictchangesauditmodel': {
            'Meta': {'object_name': 'DictChangesAuditModel', 'db_table': "'m3_audit_dict_changes'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'object_data': ('django.db.models.fields.TextField', [], {}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'db_index': 'True'}),
            'object_model': ('django.db.models.fields.CharField', [], {'max_length': '300', 'db_index': 'True'}),
            'type': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'user_fio': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '70', 'null': 'True', 'db_index': 'True', 'blank': 'True'}),
            'user_info': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '200', 'null': 'True', 'blank': 'True'}),
            'userid': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'db_index': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'default': "u''", 'max_length': '50', 'null': 'True', 'db_index': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['m3_audit']



#coding: utf-8
from setuptools import setup, find_packages

requires = []
with open('src/requires.txt', 'r') as f:
    requires.extend(f.readlines())

setup(name='m3-audit',
      version='1.1',
      url='https://bitbucket.org/barsgroup/m3_audit',
      license='MIT',
      author='BARS Group',
      author_email='bars@bars-open.ru',
      description=u'Аудит операций в прикладной системе',
      package_dir={'': 'src'},
      packages=find_packages('src'),
      install_requires=requires,
      include_package_data=True,
      classifiers=[
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Natural Language :: Russian',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 5 - Production/Stable',
      ],
      )



#coding:utf-8
'''
Created on 06.09.2011

@author: akvarats
'''

import json

from django.test import TestCase

class SomeTests(TestCase):
    '''
    Тесты для источников данных
    '''
    def test1(self):   
        pass
